@@ -1,13 +1,10 @@
-# A general class for utility functions.  Please use this for
-# functions that stand alone, without requiring a $self object,
-# Portability functions especially.
-
 # <@LICENSE>
-# Copyright 2004 Apache Software Foundation
-# 
-# Licensed under the Apache License, Version 2.0 (the "License");
-# you may not use this file except in compliance with the License.
-# You may obtain a copy of the License at
+# Licensed to the Apache Software Foundation (ASF) under one or more
+# contributor license agreements.  See the NOTICE file distributed with
+# this work for additional information regarding copyright ownership.
+# The ASF licenses this file to you under the Apache License, Version 2.0
+# (the "License"); you may not use this file except in compliance with
+# the License.  You may obtain a copy of the License at:
 # 
 #     http://www.apache.org/licenses/LICENSE-2.0
 # 
@@ -18,11 +15,33 @@
 # limitations under the License.
 # </@LICENSE>
 
+=head1 NAME
+
+Mail::SpamAssassin::Util - utility functions
+
+=head1 DESCRIPTION
+
+A general class for utility functions.  Please use this for functions that
+stand alone, without requiring a $self object, Portability functions
+especially.
+
+NOTE: The functions in this module are to be considered private.  Their API may
+change at any point, and it's expected that they'll only be used by other
+Mail::SpamAssassin modules. (TODO: we should probably revisit this if
+it's useful for plugin development.)
+
+=over 4
+
+=cut
+
 package Mail::SpamAssassin::Util;
 
 use strict;
+use warnings;
 use bytes;
 
+use Mail::SpamAssassin::Logger;
+
 use vars qw (
   @ISA @EXPORT
   $AM_TAINTED
@@ -42,6 +61,7 @@ use Time::Local;
 use Sys::Hostname (); # don't import hostname() into this namespace!
 use Fcntl;
 use POSIX (); # don't import anything unless we ask explicitly!
+use Errno qw(EEXIST);
 
 ###########################################################################
 
@@ -60,17 +80,17 @@ use constant RUNNING_ON_WINDOWS => ($^O =~ /^(?:mswin|dos|os2)/oi);
 
     clean_path_in_taint_mode();
     if ( !$displayed_path++ ) {
-      dbg("Current PATH is: ".join($Config{'path_sep'},File::Spec->path()));
+      dbg("util: current PATH is: ".join($Config{'path_sep'},File::Spec->path()));
     }
     foreach my $path (File::Spec->path()) {
       my $fname = File::Spec->catfile ($path, $filename);
       if ( -f $fname ) {
         if (-x $fname) {
-          dbg ("executable for $filename was found at $fname");
+          dbg("util: executable for $filename was found at $fname");
           return $fname;
         }
         else {
-          dbg("$filename was found at $fname, but isn't executable");
+          dbg("util: $filename was found at $fname, but isn't executable");
         }
       }
     }
@@ -86,10 +106,15 @@ use constant RUNNING_ON_WINDOWS => ($^O =~ /^(?:mswin|dos|os2)/oi);
   my $cleaned_taint_path = 0;
 
   sub clean_path_in_taint_mode {
-    return if ( $cleaned_taint_path++ );
+    return if ($cleaned_taint_path++);
     return unless am_running_in_taint_mode();
 
-    dbg("Running in taint mode, removing unsafe env vars, and resetting PATH");
+    dbg("util: taint mode: deleting unsafe environment variables, resetting PATH");
+
+    if (RUNNING_ON_WINDOWS) {
+      dbg("util: running on Win32, skipping PATH cleaning");
+      return;
+    }
 
     delete @ENV{qw(IFS CDPATH ENV BASH_ENV)};
 
@@ -103,31 +128,31 @@ use constant RUNNING_ON_WINDOWS => ($^O =~ /^(?:mswin|dos|os2)/oi);
       $dir = File::Spec->canonpath($1);
 
       if (!File::Spec->file_name_is_absolute($dir)) {
-	dbg("PATH included '$dir', which is not absolute, dropping.");
+	dbg("util: PATH included '$dir', which is not absolute, dropping");
 	next;
       }
       elsif (!(@stat=stat($dir))) {
-	dbg("PATH included '$dir', which doesn't exist, dropping.");
+	dbg("util: PATH included '$dir', which doesn't exist, dropping");
 	next;
       }
       elsif (!-d _) {
-	dbg("PATH included '$dir', which isn't a directory, dropping.");
+	dbg("util: PATH included '$dir', which isn't a directory, dropping");
 	next;
       }
       elsif (($stat[2]&2) != 0) {
         # World-Writable directories are considered insecure.
         # We could be more paranoid and check all of the parent directories as well,
         # but it's good for now.
-	dbg("PATH included '$dir', which is world writable, dropping.");
+	dbg("util: PATH included '$dir', which is world writable, dropping");
 	next;
       }
 
-      dbg("PATH included '$dir', keeping.");
+      dbg("util: PATH included '$dir', keeping");
       push(@path, $dir);
     }
 
     $ENV{'PATH'} = join($Config{'path_sep'}, @path);
-    dbg("Final PATH set to: ".$ENV{'PATH'});
+    dbg("util: final PATH set to: ".$ENV{'PATH'});
   }
 }
 
@@ -156,7 +181,7 @@ sub am_running_in_taint_mode {
     # seriously mind-bending perl
     $AM_TAINTED = not eval { eval "1 || $blank" || 1 };
   }
-  dbg ("running in taint mode? ". ($AM_TAINTED ? "yes" : "no"));
+  dbg("util: running in taint mode? ". ($AM_TAINTED ? "yes" : "no"));
   return $AM_TAINTED;
 }
 
@@ -184,13 +209,13 @@ sub untaint_file_path {
 
   # Barry Jaspan: allow ~ and spaces, good for Windows.  Also return ''
   # if input is '', as it is a safe path.
-  my $chars = '-_A-Za-z\xA0-\xFF0-9\.\@\=\+\,\/\\\:';
+  my $chars = '-_A-Za-z\xA0-\xFF0-9\.\%\@\=\+\,\/\\\:';
   my $re = qr/^\s*([$chars][${chars}~ ]*)$/o;
 
   if ($path =~ $re) {
     return $1;
   } else {
-    warn "security: cannot untaint path: \"$path\"\n";
+    warn "util: cannot untaint path: \"$path\"\n";
     return $path;
   }
 }
@@ -201,15 +226,16 @@ sub untaint_hostname {
   return unless defined($host);
   return '' if ($host eq '');
 
-  # from RFC 1035, but allowing domains starting with numbers
-  my $label = q/[A-Za-z\d](?:[A-Za-z\d-]{0,61}[A-Za-z\d])?/;
-  my $domain = qq<$label(?:\.$label)*>;
-
-  if (length($host) <= 255 && $host =~ /^($domain)$/) {
+  # from RFC 1035, but allowing domains starting with numbers:
+  #   $label = q/[A-Za-z\d](?:[A-Za-z\d-]{0,61}[A-Za-z\d])?/;
+  #   $domain = qq<$label(?:\.$label)*>;
+  #   length($host) <= 255 && $host =~ /^($domain)$/
+  # expanded (no variables in the re) because of a tainting bug in Perl 5.8.0
+  if (length($host) <= 255 && $host =~ /^([a-z\d](?:[a-z\d-]{0,61}[a-z\d])?(?:\.[a-z\d](?:[a-z\d-]{0,61}[a-z\d])?)*)$/i) {
     return $1;
   }
   else {
-    warn "security: cannot untaint hostname: \"$host\"\n";
+    warn "util: cannot untaint hostname: \"$host\"\n";
     return $host;
   }
 }
@@ -251,13 +277,27 @@ sub untaint_var {
     ${$_} = untaint_var(${$_});
   }
   else {
-    warn "Can't untaint a " . ref($_) . "!\n";
+    warn "util: can't untaint a " . ref($_) . "!\n";
   }
   return $_;
 }
 
 ###########################################################################
 
+sub taint_var {
+  my ($v) = @_;
+  return $v unless defined $v;      # can't taint "undef"
+
+  # $^X is apparently "always tainted".  We can use this to render
+  # a string tainted as follows:
+  my $tainter = substr ($^X."_", 0, 1);     # get 1 tainted char
+  $v .= $tainter; chop $v;      # then add and remove it
+
+  return $v;
+}
+
+###########################################################################
+
 # timezone mappings: in case of conflicts, use RFC 2822, then most
 # common and least conflicting mapping
 my %TZ = (
@@ -324,13 +364,18 @@ my %TZ = (
 my %MONTH = (jan => 1, feb => 2, mar => 3, apr => 4, may => 5, jun => 6,
 	     jul => 7, aug => 8, sep => 9, oct => 10, nov => 11, dec => 12);
 
+my $LOCALTZ;
+
 sub local_tz {
+  return $LOCALTZ if defined($LOCALTZ);
+
   # standard method for determining local timezone
   my $time = time;
   my @g = gmtime($time);
   my @t = localtime($time);
   my $z = $t[1]-$g[1]+($t[2]-$g[2])*60+($t[7]-$g[7])*1440+($t[5]-$g[5])*525600;
-  return sprintf("%+.2d%.2d", $z/60, $z%60);
+  $LOCALTZ = sprintf("%+.2d%.2d", $z/60, $z%60);
+  return $LOCALTZ;
 }
 
 sub parse_rfc822_date {
@@ -349,7 +394,7 @@ sub parse_rfc822_date {
   } elsif (s/ (\d+) (Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec) (\d{2,3}) / /i) {
     $dd = $1; $mon = lc($2); $yyyy = $3;
   } else {
-    dbg ("time cannot be parsed: $date");
+    dbg("util: time cannot be parsed: $date");
     return undef;
   }
 
@@ -387,15 +432,48 @@ sub parse_rfc822_date {
 
   $hh ||= 0; $mm ||= 0; $ss ||= 0; $dd ||= 0; $mmm ||= 0; $yyyy ||= 0;
 
+  # Fudge invalid times so that we get a usable date.
+  if ($ss > 59) { 
+    dbg("util: second after supported range, forcing second to 59: $date");  
+    $ss = 59;
+  } 
+
+  if ($mm > 59) { 
+    dbg("util: minute after supported range, forcing minute to 59: $date");
+    $mm = 59;
+  }
+
+  if ($hh > 23) { 
+    dbg("util: hour after supported range, forcing hour to 23: $date"); 
+    $hh = 23;
+  }
+
+  my $max_dd = 31;
+  if ($mmm == 4 || $mmm == 6 || $mmm == 9 || $mmm == 11) {
+    $max_dd = 30;
+  }
+  elsif ($mmm == 2) {
+    $max_dd = (!($yyyy % 4) && (($yyyy % 100) || !($yyyy % 400))) ? 29 : 28;
+  }
+  if ($dd > $max_dd) {
+    dbg("util: day is too high, incrementing date to next valid date: $date");
+    $dd = 1;
+    $mmm++;
+    if ($mmm > 12) {
+      $mmm = 1;
+      $yyyy++;
+    }
+  }
+
   # Time::Local (v1.10 at least) throws warnings when the dates cause
   # a 32-bit overflow.  So force a min/max for year.
   if ($yyyy > 2037) {
-    dbg("util: forcing year to 2037: $date");
-    $yyyy=2037;
+    dbg("util: year after supported range, forcing year to 2037: $date");
+    $yyyy = 2037;
   }
   elsif ($yyyy < 1970) {
-    dbg("util: forcing year to 1970: $date");
-    $yyyy=1970;
+    dbg("util: year before supported range, forcing year to 1970: $date");
+    $yyyy = 1971;
   }
 
   my $time;
@@ -404,7 +482,7 @@ sub parse_rfc822_date {
   };
 
   if ($@) {
-    dbg ("time cannot be parsed: $date, $yyyy-$mmm-$dd $hh:$mm:$ss");
+    dbg("util: time cannot be parsed: $date, $yyyy-$mmm-$dd $hh:$mm:$ss");
     return undef;
   }
 
@@ -435,11 +513,91 @@ sub time_to_rfc822_date {
 
 ###########################################################################
 
+# This used to be a wrapper for Text::Wrap.  Now we do basically the same
+# function as Text::Wrap::wrap().  See bug 5056 and 2165 for more information
+# about why things aren't using that function anymore.
+#
+# It accepts values for almost all options which can be set
+# in Text::Wrap.   All parameters are optional (leaving away the first one 
+# probably doesn't make too much sense though), either a missing or a false
+# value will fall back to the default.
+# 
+# The parameters are:
+#  1st:  The string to wrap.  Only one string is allowed.
+#                                             (default: "")
+#  2nd:  The prefix to be put in front of all lines except the first one. 
+#                                             (default: "")
+#  3rd:  The prefix for the first line.       (default:  "")
+#  4th:  The number of columns available (no line will be longer than this
+#        unless overflow is set below).       (default:  77)
+#  5th:  Enable or disable overflow mode.     (default: 0)
+#  6th:  The sequence/expression to wrap at.  (default: '\s');
+#  7th:  The string to join the lines again.  (default: "\n")
+
+sub wrap {
+  my $string   = shift || '';
+  my $prefix   = shift || '';
+  my $first    = shift || '';
+  my $length   = shift || 77;
+  my $overflow = shift || 0;
+  my $break    = shift || qr/\s/;
+  my $sep      = "\n";
+
+  # go ahead and break apart the string, keeping the break chars
+  my @arr = split(/($break)/, $string);
+
+  # tack the first prefix line at the start
+  splice @arr, 0, 0, $first if $first;
+
+  # go ahead and make up the lines in the array
+  my $pos = 0;
+  my $pos_mod = 0;
+  while ($#arr > $pos) {
+    my $len = length $arr[$pos];
+
+    # if we don't want to have lines > $length (overflow==0), we
+    # need to verify what will happen with the next line.  if we don't
+    # care if a single line goes longer, don't care about the next
+    # line.
+    # we also want this to be true for the first entry on the line
+    if ($pos_mod != 0 && $overflow == 0) {
+      $len += length $arr[$pos+1];
+    }
+
+    if ($len <= $length) {
+      # if the length determined above is within bounds, go ahead and
+      # merge the next line with the current one
+      $arr[$pos] .= splice @arr, $pos+1, 1;
+      $pos_mod = 1;
+    }
+    else {
+      # ok, the current line is the right length, but there's more text!
+      # prep the current line and then go onto the next one
+
+      # strip any trailing whitespace from the next line that's ready
+      $arr[$pos] =~ s/\s+$//;
+
+      # go to the next line and reset pos_mod
+      $pos++;
+      $pos_mod = 0;
+
+      # put the appropriate prefix at the front of the line
+      splice @arr, $pos, 0, $prefix;
+    }
+  }
+
+  # go ahead and return the wrapped text, with the separator in between
+  return join($sep, @arr);
+}
+
+###########################################################################
+
 # Some base64 decoders will remove intermediate "=" characters, others
 # will stop decoding on the first "=" character, this one translates "="
 # characters to null.
 sub base64_decode {
   local $_ = shift;
+  my $decoded_length = shift;
 
   s/\s+//g;
   if (HAS_MIME_BASE64 && (length($_) % 4 == 0) &&
@@ -447,11 +605,21 @@ sub base64_decode {
   {
     # only use MIME::Base64 when the XS and Perl are both correct and quiet
     s/(=+)(?!=*$)/'A' x length($1)/ge;
+
+    # If only a certain number of bytes are requested, truncate the encoded
+    # version down to the appropriate size and return the requested bytes
+    if (defined $decoded_length) {
+      $_ = substr $_, 0, 4 * (int($decoded_length/3) + 1);
+      my $decoded = MIME::Base64::decode_base64($_);
+      return substr $decoded, 0, $decoded_length;
+    }
+
+    # otherwise, just decode the whole thing and return it
     return MIME::Base64::decode_base64($_);
   }
-  tr|A-Za-z0-9+/=||cd;			# remove non-base64 characters
+  tr{A-Za-z0-9+/=}{}cd;			# remove non-base64 characters
   s/=+$//;				# remove terminating padding
-  tr|A-Za-z0-9+/=| -_`|;		# translate to uuencode
+  tr{A-Za-z0-9+/=}{ -_`};		# translate to uuencode
   s/.$// if (length($_) % 4 == 1);	# unpack cannot cope with extra byte
 
   my $length;
@@ -459,6 +627,13 @@ sub base64_decode {
   while ($_) {
     $length = (length >= 84) ? 84 : length;
     $out .= unpack("u", chr(32 + $length * 3/4) . substr($_, 0, $length, ''));
+    last if (defined $decoded_length && length $out >= $decoded_length);
+  }
+
+  # If only a certain number of bytes are requested, truncate the encoded
+  # version down to the appropriate size and return the requested bytes
+  if (defined $decoded_length) {
+    return substr $out, 0, $decoded_length;
   }
 
   return $out;
@@ -481,7 +656,7 @@ sub base64_encode {
 
   $_ = pack("u57", $_);
   s/^.//mg;
-  tr| -_`|A-Za-z0-9+/A|;
+  tr| -_`|A-Za-z0-9+/A|; # -> #`# <- kluge against vim syntax issues
   s/(A+)$/'=' x length $1/e;
   return $_;
 }
@@ -496,18 +671,18 @@ sub portable_getpwuid {
   if (!RUNNING_ON_WINDOWS) {
     eval ' sub _getpwuid_wrapper { getpwuid($_[0]); } ';
   } else {
-    dbg ("defining getpwuid() wrapper using 'unknown' as username");
-    eval ' sub _getpwuid_wrapper { fake_getpwuid($_[0]); } ';
+    dbg("util: defining getpwuid() wrapper using 'unknown' as username");
+    eval ' sub _getpwuid_wrapper { _fake_getpwuid($_[0]); } ';
   }
 
   if ($@) {
-    warn "Failed to define getpwuid() wrapper: $@\n";
+    warn "util: failed to define getpwuid() wrapper: $@\n";
   } else {
     return Mail::SpamAssassin::Util::_getpwuid_wrapper(@_);
   }
 }
 
-sub fake_getpwuid {
+sub _fake_getpwuid {
   return (
     'unknown',		# name,
     'x',		# passwd,
@@ -554,7 +729,7 @@ sub extract_ipv4_addr_from_string {
   my($hostname, $fq_hostname);
 
 # get the current host's unqalified domain name (better: return whatever
-# Sys::Hostname thinks out hostname is, might also be a full qualified one)
+# Sys::Hostname thinks our hostname is, might also be a full qualified one)
   sub hostname {
     return $hostname if defined($hostname);
 
@@ -668,14 +843,27 @@ sub parse_content_type {
   # Get the actual MIME type out ...
   # Note: the header content may not be whitespace unfolded, so make sure the
   # REs do /s when appropriate.
+  # correct:
+  # Content-type: text/plain; charset=us-ascii
+  # missing a semi-colon, CT shouldn't have whitespace anyway:
+  # Content-type: text/plain charset=us-ascii
   #
-  $ct =~ s/;.*$//s;                     # strip everything after first semi-colon
-  $ct =~ s@^([^/]+(?:/[^/]*)?).*$@$1@s;	# only something/something ...
-  $ct =~ tr/\000-\040\177-\377\042\050\051\054\056\072-\077\100\133-\135//d;    # strip inappropriate chars
+  $ct =~ s/^\s+//;				# strip leading whitespace
+  $ct =~ s/;.*$//s;				# strip everything after first ';'
+  $ct =~ s@^([^/]+(?:/[^/\s]*)?).*$@$1@s;	# only something/something ...
   $ct = lc $ct;
 
-  # bug 4298: If at this point we don't have a content-type, assume text/plain
-  $ct ||= "text/plain";
+  # bug 4298: If at this point we don't have a content-type, assume text/plain;
+  # also, bug 5399: if the content-type *starts* with "text", and isn't in a 
+  # list of known bad/non-plain formats, do likewise.
+  if (!$ct ||
+        ($ct =~ /^text\b/ && $ct !~ /^text\/(?:x-vcard|calendar|html)$/))
+  {
+    $ct = "text/plain";
+  }
+
+  # strip inappropriate chars (bug 5399: after the text/plain fixup)
+  $ct =~ tr/\000-\040\177-\377\042\050\051\054\056\072-\077\100\133-\135//d;
 
   # Now that the header has been parsed, return the requested information.
   # In scalar context, just the MIME type, in array context the
@@ -706,8 +894,8 @@ sub url_encode {
     }
     # other stuff
     else {
-      # 0x00-0x20, 0x7f-0xff, <, >, and " ... "
-      s/([\000-\040\177-\377\074\076\042])
+      # 0x00-0x20, 0x7f-0xff, ", %, <, >
+      s/([\000-\040\177-\377\042\045\074\076])
 	  /push(@encoded, $1) && sprintf "%%%02x", unpack("C",$1)/egx;
     }
   }
@@ -723,8 +911,8 @@ sub url_encode {
 
 =item $module = first_available_module (@module_list)
 
-Return the first module that can be successfully loaded with C<require>
-from the list.  Returns C<undef> if none are available.
+Return the name of the first module that can be successfully loaded with
+C<require> from the list.  Returns C<undef> if none are available.
 
 This is used instead of C<AnyDBM_File> as follows:
 
@@ -749,63 +937,146 @@ sub first_available_module {
 
 ###########################################################################
 
-# thanks to http://www2.picante.com:81/~gtaylor/autobuse/ for this
-# code.
+=item my ($filepath, $filehandle) = secure_tmpfile();
+
+Generates a filename for a temporary file, opens it exclusively and
+securely, and returns a filehandle to the open file (opened O_RDWR).
+
+If it cannot open a file after 20 tries, it returns C<undef>.
+
+=cut
+
+# thanks to http://www2.picante.com:81/~gtaylor/autobuse/ for this code
 sub secure_tmpfile {
   my $tmpdir = Mail::SpamAssassin::Util::untaint_file_path(
-                 File::Spec->tmpdir()
-               );
+                $ENV{'TMPDIR'} || File::Spec->tmpdir());
+
   if (!$tmpdir) {
-    die "Cannot find a temporary directory! set TMP or TMPDIR in env";
+    # Note: we would prefer to keep this fatal, as not being able to
+    # find a writable tmpdir is a big deal for the calling code too.
+    # That would be quite a psychotic case, also.
+    warn "util: cannot find a temporary directory, set TMP or TMPDIR in environment";
+    return;
   }
 
-  my ($reportfile,$tmpfile);
+  my ($reportfile, $tmpfile);
   my $umask = umask 077;
-  do {
-    # we do not rely on the obscurity of this name for security...
+
+  for (my $retries = 20; $retries > 0; $retries--) {
+    # we do not rely on the obscurity of this name for security,
     # we use a average-quality PRG since this is all we need
-    my $suffix = join ('',
-                       (0..9, 'A'..'Z','a'..'z')[rand 62,
-                                                 rand 62,
-                                                 rand 62,
-                                                 rand 62,
-                                                 rand 62,
-                                                 rand 62]);
-    $reportfile = File::Spec->catfile(
-                    $tmpdir,
-                    join ('.',
-                      "spamassassin",
-                      $$,
-                      $suffix,
-                      "tmp",
-                    )
-                  );
-    # ...rather, we require O_EXCL|O_CREAT to guarantee us proper
-    # ownership of our file; read the open(2) man page.
-  } while (! sysopen ($tmpfile, $reportfile, O_RDWR|O_CREAT|O_EXCL, 0600));
+    my $suffix = join('', (0..9,'A'..'Z','a'..'z')[rand 62, rand 62, rand 62,
+						   rand 62, rand 62, rand 62]);
+    $reportfile = File::Spec->catfile($tmpdir,".spamassassin${$}${suffix}tmp");
+
+    # instead, we require O_EXCL|O_CREAT to guarantee us proper
+    # ownership of our file, read the open(2) man page
+    if (sysopen($tmpfile, $reportfile, O_RDWR|O_CREAT|O_EXCL, 0600)) {
+      binmode $tmpfile;
+      last;
+    }
+
+    if ($!{EEXIST}) {
+      # it is acceptable if $tmpfile already exists, try another
+      next;
+    }
+    
+    # error, maybe "out of quota" or "too many open files" (bug 4017)
+    warn "util: secure_tmpfile failed to create file '$reportfile': $!\n";
+
+    # ensure the file handle is not semi-open in some way
+    if ($tmpfile) {
+      close $tmpfile;
+    }
+  }
+
   umask $umask;
 
+  if (!$tmpfile) {
+    warn "util: secure_tmpfile failed to create file, giving up";
+    return;	# undef
+  }
+
   return ($reportfile, $tmpfile);
 }
 
+=item my ($dirpath) = secure_tmpdir();
+
+Generates a directory for temporary files.  Creates it securely and
+returns the path to the directory.
+
+If it cannot create a directory after 20 tries, it returns C<undef>.
+
+=cut
+
+# stolen from secure_tmpfile()
+sub secure_tmpdir {
+  my $tmpdir = Mail::SpamAssassin::Util::untaint_file_path(File::Spec->tmpdir());
+
+  if (!$tmpdir) {
+    # Note: we would prefer to keep this fatal, as not being able to
+    # find a writable tmpdir is a big deal for the calling code too.
+    # That would be quite a psychotic case, also.
+    warn "util: cannot find a temporary directory, set TMP or TMPDIR in environment";
+    return;
+  }
+
+  my ($reportpath, $tmppath);
+  my $umask = umask 077;
+
+  for (my $retries = 20; $retries > 0; $retries--) {
+    # we do not rely on the obscurity of this name for security,
+    # we use a average-quality PRG since this is all we need
+    my $suffix = join('', (0..9,'A'..'Z','a'..'z')[rand 62, rand 62, rand 62,
+						   rand 62, rand 62, rand 62]);
+    $reportpath = File::Spec->catfile($tmpdir,".spamassassin${$}${suffix}tmp");
+
+    # instead, we require O_EXCL|O_CREAT to guarantee us proper
+    # ownership of our file, read the open(2) man page
+    if (mkdir $reportpath, 0700) {
+      $tmppath = $reportpath;
+      last;
+    }
+
+    if ($!{EEXIST}) {
+      # it is acceptable if $reportpath already exists, try another
+      next;
+    }
+    
+    # error, maybe "out of quota" or "too many open files" (bug 4017)
+    warn "util: secure_tmpdir failed to create file '$reportpath': $!\n";
+  }
+
+  umask $umask;
+
+  warn "util: secure_tmpdir failed to create a directory, giving up" if (!$tmppath);
+
+  return $tmppath;
+}
+
+
 ###########################################################################
 
 sub uri_to_domain {
   my ($uri) = @_;
 
-  #return if ($uri =~ /^mailto:/i);	# not mailto's, please (TODO?)
-
   # Javascript is not going to help us, so return.
   return if ($uri =~ /^javascript:/i);
 
   $uri =~ s,#.*$,,gs;			# drop fragment
   $uri =~ s#^[a-z]+:/{0,2}##gsi;	# drop the protocol
   $uri =~ s,^[^/]*\@,,gs;		# username/passwd
-  $uri =~ s,[/\?].*$,,gs;		# path/cgi params
+
+  # strip path and CGI params.  note: bug 4213 shows that "&" should
+  # *not* be likewise stripped here -- it's permitted in hostnames by
+  # some common MUAs!
+  $uri =~ s,[/\?].*$,,gs;              
+
   $uri =~ s,:\d*$,,gs;			# port, bug 4191: sometimes the # is missing
 
-  return if $uri =~ /\%/;         # skip undecoded URIs.
-  # we'll see the decoded version as well
+  # skip undecoded URIs if the encoded bits shouldn't be.
+  # we'll see the decoded version as well.  see url_encode()
+  return if $uri =~ /\%(?:2[1-9a-fA-F]|[3-6][0-9a-fA-f]|7[0-9a-eA-E])/;
 
   # keep IPs intact
   if ($uri !~ /^\d+\.\d+\.\d+\.\d+$/) { 
@@ -822,13 +1093,14 @@ sub uri_to_domain {
 }
 
 sub uri_list_canonify {
-  my(@uris) = @_;
+  my($redirector_patterns, @uris) = @_;
 
   # make sure we catch bad encoding tricks
   my @nuris = ();
   for my $uri (@uris) {
-    # we're interested in http:// and so on, skip mailto:
-    next if $uri =~ /^mailto:/i;
+    # we're interested in http:// and so on, skip mailto: and
+    # email addresses with no protocol
+    next if $uri =~ /^mailto:/i || $uri =~ /^[^:]*\@/;
 
     # sometimes we catch URLs on multiple lines
     $uri =~ s/\n//g;
@@ -843,6 +1115,11 @@ sub uri_list_canonify {
     # Make a copy so we don't trash the original in the array
     my $nuri = $uri;
 
+    # bug 4390: certain MUAs treat back slashes as front slashes.
+    # since backslashes are supposed to be encoded in a URI, swap non-encoded
+    # ones with front slashes.
+    $nuri =~ tr@\\@/@;
+
     # http:www.foo.biz -> http://www.foo.biz
     $nuri =~ s#^(https?:)/{0,2}#$1//#i;
 
@@ -878,7 +1155,7 @@ sub uri_list_canonify {
     }
 
     # deal with wierd hostname parts, remove user/pass, etc.
-    if ($nuri =~ m{^(https?://)([^/]+)(\/.*)?$}i) {
+    if ($nuri =~ m{^(https?://)([^/]+?)((?::\d*)?\/.*)?$}i) {
       my($proto, $host, $rest) = ($1,$2,$3);
 
       # not required
@@ -899,6 +1176,20 @@ sub uri_list_canonify {
         push(@uris, $1);
       }
 
+      # resort to redirector pattern matching if the generic https? check
+      # doesn't result in a match -- bug 4176
+      else {
+	foreach (@{$redirector_patterns}) {
+	  if ("$proto$host$rest" =~ $_) {
+	    next unless defined $1;
+	    dbg("uri: parsed uri pattern: $_");
+	    dbg("uri: parsed uri found: $1 in redirector: $proto$host$rest");
+	    push (@uris, $1);
+	    last;
+	  }
+	}
+      }
+
       ########################
       ## TVD: known issue, if host has multiple combinations of the following,
       ## all permutations will be put onto @nuris.  shouldn't be an issue.
@@ -920,11 +1211,23 @@ sub uri_list_canonify {
 
       ########################
 
-      # deal with 'http://213.172.0x1f.13/', decode encoded octets
-      if ($host =~ /^([0-9a-fx]*\.)([0-9a-fx]*\.)([0-9a-fx]*\.)([0-9a-fx]*)$/ix) {
-        my (@chunk) = ($1,$2,$3,$4);
-        for my $octet (0 .. 3) {
-          $chunk[$octet] =~ s/^0x([0-9a-f][0-9a-f])/sprintf "%d",hex($1)/gei;
+      # deal with hosts which are IPs
+      # also handle things like:
+      # http://89.0x00000000000000000000068.0000000000000000000000160.0x00000000000011
+      #    both hex (0x) and oct (0+) encoded octets, etc.
+
+      if ($host =~ /^
+                    ((?:0x[0-9a-f]+|\d+)\.)
+                    ((?:0x[0-9a-f]+|\d+)\.)
+                    ((?:0x[0-9a-f]+|\d+)\.)
+                    (0x[0-9a-f]+|\d+)
+                    $/ix)
+      {
+        my @chunk = ($1,$2,$3,$4);
+        foreach my $octet (@chunk) {
+          $octet =~ s/^0x([0-9a-f]+)/sprintf "%d",hex($1)/gei;
+          $octet =~ s/^0+([1-3][0-7]{0,2}|[4-7][0-7]?)\b/sprintf "%d",oct($1)/ge;
+	  $octet =~ s/^0+//;
         }
         push(@nuris, join ('', $proto, @chunk, $rest));
       }
@@ -1029,7 +1332,7 @@ sub setuid_to_euid {
   my $touid = $>;
 
   if ($< != $touid) {
-    dbg ("changing real uid from $< to match effective uid $touid");
+    dbg("util: changing real uid from $< to match effective uid $touid");
     $< = $touid; # try the simple method first
 
     # bug 3586: Some perl versions, typically those on a BSD-based
@@ -1037,7 +1340,7 @@ sub setuid_to_euid {
     # can be changed.  So this is a kluge for us to get around the
     # typical spamd-ish behavior of: $< = 0, $> = someuid ...
     if ( $< != $touid ) {
-      dbg("initial attempt to change real uid failed, trying BSD workaround");
+      dbg("util: initial attempt to change real uid failed, trying BSD workaround");
 
       $> = $<;			# revert euid to ruid
       $< = $touid;		# change ruid to target
@@ -1046,7 +1349,8 @@ sub setuid_to_euid {
 
     # Check that we have now accomplished the setuid
     if ($< != $touid) {
-      die "setuid $< to $touid failed!";
+      # keep this fatal: it's a serious security problem if it fails
+      die "util: setuid $< to $touid failed!";
     }
   }
 }
@@ -1065,27 +1369,44 @@ sub helper_app_pipe_open_windows {
 
   # use a traditional open(FOO, "cmd |")
   my $cmd = join(' ', @cmdline);
-  if ($stdinfile) { $cmd .= " < '$stdinfile'"; }
+  if ($stdinfile) { $cmd .= qq/ < "$stdinfile"/; }
   if ($duperr2out) { $cmd .= " 2>&1"; }
   return open ($fh, $cmd.'|');
 }
 
+sub force_die {
+  my ($msg) = @_;
+
+  # note use of eval { } scope in logging -- paranoia to ensure that a broken
+  # $SIG{__WARN__} implementation will not interfere with the flow of control
+  # here, where we *have* to die.
+  eval { warn $msg; };
+
+  POSIX::_exit(1);  # avoid END and destructor processing 
+  kill('KILL',$$);  # still kicking? die! 
+}
+
 sub helper_app_pipe_open_unix {
   my ($fh, $stdinfile, $duperr2out, @cmdline) = @_;
 
   # do a fork-open, so we can setuid() back
   my $pid = open ($fh, '-|');
   if (!defined $pid) {
-    die "cannot fork: $!";
+    # acceptable to die() here, calling code catches it
+    die "util: cannot fork: $!";
   }
 
   if ($pid != 0) {
     return $pid;          # parent process; return the child pid
   }
 
-  # else, child process.  go setuid...
+  # else, child process.  
+  # from now on, we cannot die(), as a parent-process eval { } scope
+  # could intercept it! use force_die() instead  (bug 4370, cmt 2)
+
+  # go setuid...
   setuid_to_euid();
-  dbg ("setuid: helper proc $$: ruid=$< euid=$>");
+  dbg("util: setuid: ruid=$< euid=$>");
 
   # now set up the fds.  due to some wierdness, we may have to ensure that we
   # *really* close the correct fd number, since some other code may have
@@ -1105,14 +1426,15 @@ sub helper_app_pipe_open_unix {
   if ($f != 0) {
     POSIX::close(0);
   }
-  open STDIN, "<$stdinfile" or die "cannot open $stdinfile: $!";
+
+  open (STDIN, "<$stdinfile") or force_die "util: cannot open $stdinfile: $!";
 
   # this should be impossible; if we just closed fd 0, UNIX
   # fd behaviour dictates that the next fd opened (the new STDIN)
   # will be the lowest unused fd number, which should be 0.
   # so die with a useful error if this somehow isn't the case.
   if (fileno(STDIN) != 0) {
-    die "setuid: oops: fileno(STDIN) [".fileno(STDIN)."] != 0";
+    force_die "util: setuid: oops: fileno(STDIN) [".fileno(STDIN)."] != 0";
   }
 
   # ensure STDOUT is open.  since we just created a pipe to ensure this, it has
@@ -1133,20 +1455,113 @@ sub helper_app_pipe_open_unix {
     if ($f != 2) {
       POSIX::close(2);
     }
-    open STDERR, ">&STDOUT" or die "dup STDOUT failed: $!";
+
+    open (STDERR, ">&STDOUT") or force_die "util: dup STDOUT failed: $!";
 
     # STDERR must be fd 2 to be useful to subprocesses! (bug 3649)
     if (fileno(STDERR) != 2) {
-      die "setuid: oops: fileno(STDERR) [".fileno(STDERR)."] != 2";
+      force_die "util: oops: fileno(STDERR) [".fileno(STDERR)."] != 2";
     }
   }
 
   exec @cmdline;
-  die "exec failed: $!";
+  warn "util: exec failed: $!";
+
+  # bug 4370: we really have to exit here; break any eval traps
+  POSIX::_exit(1);  # avoid END and destructor processing 
+  kill('KILL',$$);  # still kicking? die! 
+  die;  # must be a die() otherwise -w will complain
 }
 
 ###########################################################################
 
-sub dbg { Mail::SpamAssassin::dbg (@_); }
+# As "perldoc perlvar" notes, in perl 5.8.0, the concept of "safe" signal
+# handling was added, which means that signals cannot interrupt a running OP.
+# unfortunately, a regexp match is a single OP, so a psychotic m// can
+# effectively "hang" the interpreter as a result, and a $SIG{ALRM} handler
+# will never get called.
+#
+# However, by using "unsafe" signals, we can still interrupt that -- and
+# POSIX::sigaction can create an unsafe handler on 5.8.x.   So this function
+# provides a portable way to do that.
+
+sub trap_sigalrm_fully {
+  my ($handler) = @_;
+  if ($] < 5.008) {
+    # signals are always unsafe, just use %SIG
+    $SIG{ALRM} = $handler;
+  } else {
+    # may be using "safe" signals with %SIG; use POSIX to avoid it
+    POSIX::sigaction POSIX::SIGALRM(), new POSIX::SigAction $handler;
+  }
+}
+
+###########################################################################
+
+# Removes any normal perl-style regexp delimiters at
+# the start and end, and modifiers at the end (if present).
+# If modifiers are found, they are inserted into the pattern using
+# the /(?i)/ idiom.
+
+sub regexp_remove_delimiters {
+  my ($re) = @_;
+
+  my $delim;
+  if (!defined $re || $re eq '') {
+    warn "cannot remove delimiters from null regexp";
+    return undef;   # invalid
+  }
+  elsif ($re =~ s/^m{//) {              # m{foo/bar}
+    $delim = '}';
+  }
+  elsif ($re =~ s/^m\(//) {             # m(foo/bar)
+    $delim = ')';
+  }
+  elsif ($re =~ s/^m<//) {              # m<foo/bar>
+    $delim = '>';
+  }
+  elsif ($re =~ s/^m(\W)//) {           # m#foo/bar#
+    $delim = $1;
+  } else {                              # /foo\/bar/ or !foo/bar!
+    $re =~ s/^(\W)//; $delim = $1;
+  }
+
+  $re =~ s/\Q${delim}\E([imsx]*)$// or warn "unbalanced re: $re";
+
+  my $mods = $1;
+  if ($mods) {
+    $re = "(?".$mods.")".$re;
+  }
+
+  return $re;
+}
+
+# turn "/foobar/i" into qr/(?i)foobar/
+
+sub make_qr {
+  my ($re) = @_;
+  $re = regexp_remove_delimiters($re);
+  return qr/$re/;
+}
+
+###########################################################################
+
+sub get_my_locales {
+  my ($ok_locales) = @_;
+
+  my @locales = split(' ', $ok_locales);
+  my $lang = $ENV{'LC_ALL'};
+  $lang ||= $ENV{'LANGUAGE'};
+  $lang ||= $ENV{'LC_MESSAGES'};
+  $lang ||= $ENV{'LANG'};
+  push (@locales, $lang) if defined($lang);
+  return @locales;
+}
+
+###########################################################################
 
 1;
+
+=back
+
+=cut
